////////////////////////////////////////////////////////////////////////////
//
// Copyright (C) DSTC Pty Ltd (ACN 052 372 577) 1993, 1994, 1995.
// Unpublished work.  All Rights Reserved.
//
// The software contained on this media is the property of the
// DSTC Pty Ltd.  Use of this software is strictly in accordance
// with the license agreement in the accompanying LICENSE.DOC
// file. If your distribution of this software does not contain
// a LICENSE.DOC file then you have no rights to use this
// software in any manner and should contact DSTC at the address
// below to determine an appropriate licensing arrangement.
//
//      DSTC Pty Ltd
//      Level 7, GP South
//      University of Queensland
//      St Lucia, 4072
//      Australia
//      Tel: +61 7 3365 4310
//      Fax: +61 7 3365 4311
//      Email: jcsi@dstc.qut.edu.au
//
// This software is being provided "AS IS" without warranty of
// any kind.  In no event shall DSTC Pty Ltd be liable for
// damage of any kind arising out of or in connection with
// the use or performance of this software.
//
////////////////////////////////////////////////////////////////////////////

package com.dstc.security.ssl;

import java.util.Iterator;
import java.util.Vector;
import java.util.HashSet;
import java.util.Properties;
import java.net.Socket;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.io.IOException;
import java.io.File;
import java.io.FileInputStream;
import java.io.ByteArrayInputStream;
import java.io.DataInputStream;
import java.io.InputStreamReader;
import java.io.BufferedReader;

import java.security.SecureRandom;
import java.security.PrivateKey;
import java.security.cert.CertificateFactory;
import java.security.cert.X509Certificate;

import javax.net.SocketFactory;
import javax.net.ssl.SSLKeyException;

import com.dstc.security.pkcs8.PrivateKeyInfo;
import com.dstc.security.pkcs8.EncryptedPrivateKeyInfo;
import com.dstc.security.pkcs8.PKCS8EncryptedPrivateKey;
import com.dstc.security.util.Config;

import fi.hut.tcm.spki.*;
import fi.hut.tcm.spki.impl.*;
import java.security.Provider;
import java.security.Security;

/**
 * JCSI's  implementation of javax.net.ssl.SSLSocketFactory
 *
 * @author Ming Yung
 */
public final class SSLSocketFactory extends BaseSSLSocketFactory
{
  private CertificateFactory certFact;
  private String keyDir = null;
  private static final String SSL_CONF = "jcsi.ssl.conf";
  private static final String SSL_PROPERTIES = "ssl.properties";

  static
  {
    if (instance == null)
    {
      try
      {
        instance = new SSLSocketFactory();
      }
      catch (IOException e)
      {
        e.printStackTrace();
      }
    }
  }

  public static SocketFactory getDefault()
  {
    return instance;
  }

  protected SSLSocketFactory() throws SSLKeyException
  {
    super();
    initializeKeys();
  }

  protected void initializeKeys() throws SSLKeyException
  {
    try
    {

      Security.addProvider(new SPKIProvider());

      certFact = CertificateFactory.getInstance("SPKI");

      Properties props = Config.getProperties(SSL_CONF, SSL_PROPERTIES);
      keyDir = (String)props.get("jcsi.ssl.key.dir");

      String debugLevel = (String)props.get("jcsi.ssl.debugLevel");
      Debug.debug = Integer.parseInt(debugLevel);

      Vector suites = new Vector();
      for (int i=1; ; i++)
      {
        String suite = (String)props.get("jcsi.ssl.enabledCipherSuite." + i);
        if (suite == null)
          break;
        suites.addElement(suite);
      }
      String[] enabledCipherSuites = new String[suites.size()];
      suites.toArray(enabledCipherSuites);
      setDefaultCipherSuites(enabledCipherSuites);

      loadKeyAndCerts(props);
      loadTrustedCerts(props);
    }
    catch (Exception e)
    {
      e.printStackTrace();
    }
  }

  private void loadKeyAndCerts(Properties props)
    throws IOException
  {
    try
    {
      String clientPrivKey = (String)props.get("jcsi.ssl.clientPrivKey");
      if (clientPrivKey == null)
        return;

      FileInputStream fis
        = new FileInputStream(keyDir + File.separator + clientPrivKey);
      byte[] encoded = new byte[fis.available()];
      fis.read(encoded);
      fis.close();

      String password
        = (String)props.get("jcsi.ssl.clientPrivKey.password");
      if (password == null)
      {
        System.out.print("Enter password to unlock private key: ");
        InputStreamReader isr = new InputStreamReader(System.in);
        BufferedReader in = new BufferedReader(isr);
        password = in.readLine();
      }

      PKCS8EncryptedPrivateKey encKey = new PKCS8EncryptedPrivateKey(encoded);
      encKey.decrypt(password);
      setClientPrivateKey(encKey.getPrivateKey());

      String certs = (String)props.get("jcsi.ssl.clientCertsChain");
      if (certs == null)
        return;

      Vector clientCerts = new Vector();
      FileInputStream fistr;
      File file = new File(certs);
        for (int i = 0; i < file.list().length; i++) {
           fistr = new FileInputStream(certs + File.separator + "clientcert"+ i +".spki");
           SPKICertificate cert
           =  (SPKICertificate)certFact.generateCertificate(fistr);
           fistr.close();
           clientCerts.addElement(cert);
          }

      setClientCertificates(clientCerts);

    }
    catch (Exception e)
    {
      e.printStackTrace();
    }
  }

  private void loadTrustedCerts(Properties props) throws SSLKeyException
  {
    try
    {
      String fileName = (String)props.get("jcsi.ssl.trustedCerts");
      if (fileName == null)
        return;

      File file = new File(keyDir + File.separator + fileName);
      if (!file.exists())
        file.createNewFile();

      FileInputStream is  = new FileInputStream(file);
      DataInputStream dis = new DataInputStream(is);
      byte[] certData = new byte[dis.available()];
      dis.readFully(certData);
      is.close();

      ByteArrayInputStream bais = new ByteArrayInputStream(certData);
      Vector trustedCerts = new Vector();
      SPKICertificate cert = (SPKICertificate)certFact.generateCertificate(bais);
      trustedCerts.add(cert);

      setTrustedCertificates(trustedCerts);
    }
    catch (Exception e)
    {
      e.printStackTrace();
    }
  }
}
